# -*- coding: utf-8 -*-
"""
from tempfile import NamedTemporaryFile
"""

import argparse
import sys

from . import __version__
from .base import Base
from .config import Config
from .ican import Ican
from .log import logger
from .exceptions import IcanException


# ===============================
#
#  CLI Class
#
# ===============================


class CLI(Base):
    usage = """ican <COMMAND> [<ARGS>]

commands:
  bump [PART] [PRE]  increment version [minor, major, patch, prerelease, build]
                     if prerelease use PRE to set [alpha, beta, rc, dev]
  show [STYLE]       show version [semantic, public, pep440, git]
  run [PIPELINE]     run the specified PIPELINE
  rollback           restore the previous version
  init               initialize a config in the current directory
"""

    def __init__(self):
        self._register_excepthook()
        if "--version" not in sys.argv:
            self._arg_pop()
            self.config = Config()
            self.config.pre_parse()
            self._substitute_aliases()

        parser = argparse.ArgumentParser(usage=CLI.usage, prog="ican")
        parser.add_argument(
            "--dry_run", help="do not write any files", action="store_true"
        )
        parser.add_argument(
            "--verbose", help="display debug information", action="store_true"
        )
        parser.add_argument("command", help=argparse.SUPPRESS)
        parser.add_argument(
            "--version",
            action="version",
            help="display ican version",
            version=f"ican v{__version__}",
        )

        args = parser.parse_args(sys.argv[1:2])
        if not hasattr(self, args.command):
            logger.error("Unrecognized command")
            parser.print_help()
            exit(1)

        getattr(self, args.command)()
        return

    def _arg_pop(self):
        """Here we will pop --verbose and --dry_run out asap,
        that way logging can be setup before we parse the config
        file, etc.
        """

        verbose = False
        dry_run = False
        for i in range(1, len(sys.argv)):
            if sys.argv[i] == "--verbose":
                verbose = True
                sys.argv.pop(i)
                break
        for i in range(1, len(sys.argv)):
            if sys.argv[i] == "--dry_run":
                dry_run = True
                sys.argv.pop(i)
                break
        logger.setup(verbose, dry_run)
        return

    def _substitute_aliases(self):
        """At this point we have parsed the config and know any
        user-defined aliases.  We need to substitute them into
        sys.argv for a seamless alias experience.
        """

        aliases = self.config.aliases
        # if config.alias was used, insert it into sys.argv
        if len(sys.argv) < 2:
            # nothing to substitute here because no command
            return
        command = sys.argv[1]
        if aliases.get(command):
            built_in = aliases.get(command)
            sys.argv.pop(1)  # delete the alias command
            sys.argv[1:1] = built_in
        return

    def _register_excepthook(self):
        """Register our custom exception handler"""

        self._original_excepthook = sys.excepthook
        sys.excepthook = self._excepthook
        return

    def _excepthook(self, type, value, tracekback, debug=False):
        """Custom exception handler"""

        if isinstance(value, IcanException):
            if value.msg:
                value.output_method(value.msg)
            if value.e:
                for line in value.e:
                    value.output_method(line)
            if debug:
                self._original_excepthook(type, value, tracekback)
            exit_code = value.exit_code
            sys.exit(exit_code)
        else:
            self._original_excepthook(type, value, tracekback)

    def bump(self):
        """dispatched here with command bump"""

        parser = argparse.ArgumentParser(
            description="PART choices [major, minor, patch, prerelease, build]",
            usage="ican bump [PART]",
        )
        parser.add_argument(
            "part",
            nargs="?",
            default="build",
            choices=["major", "minor", "patch", "prerelease", "build"],
            help=argparse.SUPPRESS,
        )
        parser.add_argument(
            "--pre",
            type=str,
            help="set the prerelease string [alpha, beta, rc, dev]",
        )
        parser.add_argument(
            "--dry_run", help="do not write any files", action="store_true"
        )
        parser.add_argument(
            "--verbose", help="display debug information", action="store_true"
        )
        args = parser.parse_args(sys.argv[2:])

        i = Ican()
        i.bump(args.part.lower(), args.pre)
        logger.verbose("bump() COMPLETE")
        logger.info(f"Version: {i.version.semantic}")

        return

    def show(self):
        """dispatched here with command show"""

        parser = argparse.ArgumentParser(
            description="STYLE choices [semantic, public, pep440, git]",
            usage="ican show [STYLE]",
        )
        parser.add_argument(
            "style",
            nargs="?",
            default="semantic",
            choices=["semantic", "public", "pep440", "git"],
            help=argparse.SUPPRESS,
        )
        # add --verbose only to be included in --help
        parser.add_argument(
            "--verbose", help="display debug information", action="store_true"
        )
        args = parser.parse_args(sys.argv[2:])

        i = Ican()
        v = i.show(args.style)
        logger.info(f"Current {args.style} version: {v}")

        return

    def rollback(self):
        """in case of emergency, restore the previously
        persisted version.
        """

        parser = argparse.ArgumentParser(usage="ican rollback")
        # add --dry_run and --verbose only to be included in --help
        parser.add_argument(
            "--dry_run", help="do not write any files", action="store_true"
        )
        parser.add_argument(
            "--verbose", help="display debug information", action="store_true"
        )
        parser.parse_args(sys.argv[2:])

        i = Ican()
        i.rollback()
        logger.verbose("rollback() COMPLETE")
        logger.info(f"Rollback: {i.version.semantic}")

    def init(self):
        """dispatched here with command init"""

        parser = argparse.ArgumentParser(usage="ican init")
        # add --dry_run and --verbose only to be included in --help
        parser.add_argument(
            "--dry_run", help="do not write any files", action="store_true"
        )
        parser.add_argument(
            "--verbose", help="display debug information", action="store_true"
        )
        parser.parse_args(sys.argv[2:])

        self.config = Config(init=True).parse()
        logger.info("init COMPLETE")

        return

    def run(self):
        """dispatched here with command init"""

        parser = argparse.ArgumentParser(
            description="PIPELINE can be any pipeline defined in your .ican file.",
            usage="ican run [PIPELINE]",
        )
        parser.add_argument("pipeline", help=argparse.SUPPRESS)
        # add --verbose only to be included in --help
        parser.add_argument(
            "--dry_run", help="do not write any files", action="store_true"
        )
        parser.add_argument(
            "--verbose", help="display debug information", action="store_true"
        )
        args = parser.parse_args(sys.argv[2:])

        i = Ican()
        i.run_pipeline(args.pipeline)

    def test(self):
        """dispatched here with command test"""

        parser = argparse.ArgumentParser(usage="ican test [ARGS]")
        # add --dry-run and --verbose only to be included in --help
        parser.add_argument(
            "--dry-run", help="do not write any files", action="store_true"
        )
        parser.add_argument(
            "--verbose", help="display debug information", action="store_true"
        )
        parser.add_argument("first", nargs="?", help=argparse.SUPPRESS)
        args = parser.parse_args(sys.argv[2:])
        logger.verbose("verbose")
        print(f"10-4 with arg {args.first}")


def entry():
    CLI()



# -*- coding: utf-8 -*-

from .base import Base
from .version import Version
from .git import Git
from .log import logger
from .exceptions import GitDescribeError
from .exceptions import NoConfigFound
from .exceptions import RollbackNotPossible

#######################################
#
#   Bump Class
#
#######################################


class Ican(Base):
    """
    Object which will orchestrate entire program
    """

    def __init__(self):
        """Typically ican will be instantiated by cli with a half parsed
        config.  We pre-parse so logging can begin.
        """
        self.ready = False

        # make sure the config is fully parsed
        if not self.config.pre_parsed:
            self.config.parse()
        elif not self.config.parsed:
            self.config.parse()
        # Here if still config not ready, it will never be ready
        if not self.config.config_file:
            raise NoConfigFound()

        # Now config is parsed.  We can parse from config
        self.version = Version.parse(self.config.current_version)
        logger.verbose(f"discovered {self.version.semantic} @ CONFIG.version")

        # Git init
        self.git = Git()

        try:
            self.version._git_metadata = self.git.describe()
        except GitDescribeError as e:
            logger.verbose(e)
            logger.verbose("Git-versions are disabled. Does this repo have a tag?")
            self.git.disable()
        else:
            logger.verbose(f"Discovered {self.version.git} @ GIT.version")
        return

    def show(self, style):
        """
        Show the <STYLE> version
        """

        v = getattr(self.version, style)
        if v is None:
            return f"Version STYLE: {style} not available"
        return v

    def rollback(self):
        """When all else fails, this should bring the version back
        to your prior saved version.  It will also update all source
        files you have configured.
        """
        if not self.config.previous_version:
            raise RollbackNotPossible()

        # delete old, create new self.version
        del self.version
        self.version = Version.parse(self.config.previous_version)

        # Update the source files
        for file in self.config.source_files:
            file.update(self.version)

        # Now that everything else is finished, persist version
        self.config.persist_version(self.config.previous_version)

    def bump(self, part, pre):
        """This is pretty much the full process"""
        if pre and part == "prerelease":
            logger.verbose(f"Setting prerelease string to {pre}")
        if pre and part != "prerelease":
            print(pre)
            logger.warning(f"Disregarding --pre with part set to {part.upper()}.")
        logger.verbose(f"Beginning bump of <{part.upper()}>")

        self.version.bump(part, pre)
        logger.verbose(f"New value of <{part.upper()}> - {getattr(self.version, part)}")

        # Update the user's files with new version
        for file in self.config.source_files:
            file.update(self.version)

        # Run the appropriate pipeline
        # new.release, new.prerelease rebuild.release, rebuild.prerelease
        if self.version.stage:
            self.run_pipeline(self.version.stage, True)

        # Once all else is successful, persist the new version
        self.config.persist_version(self.version.semantic)

        return self

    def run_pipeline(self, pipeline, automated=False):
        # Pipeline
        if self.config.pipelines.get(pipeline) is None:
            # Pipeline is not defined
            if automated:
                logger.verbose(f"Pipeline `{pipeline}` not found.")
            else:
                # This was requested so an error is appropriate
                logger.error(f"Pipeline `{pipeline}` not found.")
            return

        pl = self.config.pipelines.get(pipeline)
        pl.run()
        return



# -*- coding: utf-8 -*-

#
#   ,---.  ,-.,---.  ,---.  ,-.    ,-..-. .-.,---.
#   | .-.\ |(|| .-.\ | .-'  | |    |(||  \| || .-'
#   | |-' )(_)| |-' )| `-.  | |    (_)|   | || `-.
#   | |--' | || |--' | .-'  | |    | || |\  || .-'
#   | |    | || |    |  `--.| `--. | || | |)||  `--.
#   /(     `-'/(     /( __.'|( __.'`-'/(  (_)/( __.'
#  (__)      (__)   (__)    (_)      (__)   (__)
#

import os
import re
import subprocess
import shlex
from types import SimpleNamespace

from .log import logger
from .base import Base


#######################################
#
#   Pipeline
#
#######################################


class Pipeline(Base):

    TEMPLATE = r"{{(?P<var>.*?)}}"

    def __init__(self, label=None, steps=None):
        self.label = label
        self.steps = []
        self.env = None
        self.ctx = None
        self.compiled = re.compile(Pipeline.TEMPLATE)

        if steps is None:
            logger.error("must include at least 1 step")

        if steps:
            for k, v in steps:
                logger.verbose(f"{label.upper()}.{k} - {v}")
                step = SimpleNamespace(label=k, cmd=v)
                self.steps.append(step)

    def _render(self, cmd):
        """render jinja-style templates
        {{var}} = ctx['var']
        """

        result, n = self.compiled.subn(
            lambda m: self.ctx.get(m.group("var").upper(), "N/A"), cmd
        )

        if n > 0:
            logger.verbose(f"rendered cmd: {result}")
        return result

    def _run_cmd(self, cmd):
        """Here is where we actually run the pipeline steps via the
        shell.

        Args:
            cmd: This should be a tuple or list of command, args such as:
            ['git', 'commit', '-a']

        Returns:
            result: the result object will have attributes of both
            stdout and stderr representing the results of the subprocess
        """

        if type(cmd) not in (tuple, list):
            cmd = shlex.split(cmd)

        logger.verbose(f"running cmd - {cmd}")
        result = subprocess.run(
            cmd, shell=False, env=self.env, capture_output=False, text=True
        ).stdout

        if result:
            logger.verbose(f"cmd result - {result}")
        return

    def _build_ctx(self):
        """ """
        self.ctx = {}
        self.ctx["VERSION"] = self.version.semantic
        self.ctx["SEMANTIC"] = self.version.semantic
        self.ctx["PUBLIC"] = self.version.public
        self.ctx["PEP440"] = self.version.pep440
        self.ctx["GIT"] = self.version.git
        self.ctx["TAG"] = self.version.tag
        self.ctx["MAJOR"] = self.version.major
        self.ctx["MINOR"] = self.version.minor
        self.ctx["PATCH"] = self.version.patch
        self.ctx["PRERELEASE"] = self.version.prerelease
        self.ctx["BUILD"] = self.version.build
        self.ctx["STAGE"] = self.version.stage
        self.ctx["ENV"] = self.version.env
        self.ctx["AGE"] = self.version.age
        self.ctx["ROOT"] = self.config.path
        self.ctx["PREVIOUS"] = self.config.previous_version

        # ensure all are strings
        for k, v in self.ctx.items():
            self.ctx[k] = str(v)

        logger.verbose(f"Generated ctx: {self.ctx}")
        return

    def _build_env(self):
        """ Use ctx and simple add prefix to all keys
        """

        _env = {f'ICAN_{k}': v for k, v in self.ctx.items()}
        self.env = {**os.environ, **_env}
        return

    def run(self):
        self._build_ctx()
        self._build_env()
        logger.info(f"+BEGIN pipeline.{self.label.upper()}")
        for step in self.steps:
            cmd = self._render(step.cmd)
            # label = step.label
            if logger.ok_to_write:
                self._run_cmd(cmd)
        logger.info(f"+END pipeline.{self.label.upper()}")
